/** 
 * @file equivalence-check.hpp
 *
 * @brief SAT-based equivalence check (respects garbage outputs and constant inputs)
 */
#include "equivalence_check.hpp" 

#include <core/circuit.hpp>
#include <core/utils/timer.hpp>
#include <core/target_tags.hpp>
#include <core/properties.hpp> 
#include <algorithms/verification/verification.hpp>

// STL and TR1
#include <tr1/tuple>
#include <utility>
#include <vector>

#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/bind.hpp>

#include <fmi/default_solver.hpp>
#include <fmi/assertions.hpp> 
#include <fmi/solution.hpp> 
#include <fmi/core.hpp>
#include <fmi/vars.hpp> 
#include <fmi/reduce.hpp>
#include <fmi/reduce-tags.hpp>
#include <fmi/integral_wrapper.hpp> // for index
#include <fmi/extract.hpp>
#include <fmi/constant.hpp>

#define foreach BOOST_FOREACH

namespace revkit
{
  using namespace std;

  namespace binary 
  {
    using namespace boost::assign; 

    fmi::bv model_toffoli( fmi::default_solver& solver, const gate& g, const fmi::bv& input )
    {    
      assert( is_toffoli( g ) && "Not a toffoli gate" );
      unsigned n = bitsize( input );

      fmi::bv output = fmi::new_variable( solver, n );

      gate::line target = *g.begin_targets();

      vector<fmi::bv> controls;
      for ( gate::const_iterator iter = g.begin_controls(); iter != g.end_controls(); ++iter )
      {
        controls += fmi::extract( solver, input, *iter );
      }

      // create identity of all lines unless target line 
      for ( unsigned i = 0; i < n; ++i )
      {
        fmi::bv outExtr = fmi::extract (solver, output, i);
        fmi::bv inExt   = fmi::extract (solver, input, i);

        if (target != i)
        {
          fmi::assertion ( solver,
              fmi::_0 == fmi::_1, outExtr, inExt);
        }
      }

      fmi::bv outT = fmi::extract (solver, output, target);
      fmi::bv inT  = fmi::extract (solver, input,  target);

      fmi::assertion (solver, 
          fmi::_0 == (fmi::_1 ^ fmi::_2),
          outT, inT,  fmi::reduce<fmi::and_op>( solver, controls.begin(), controls.end() ) );

      return output;
    }

    fmi::bv model_gate (fmi::default_solver& solver, const gate& g, const fmi::bv& input) 
    {
      assert( !solver.empty() && "No solver specified" );

      if ( is_toffoli( g ) )
      {
        return model_toffoli( solver, g, input );
      }

      std::cerr << "Gate type other than toffoli not supported. Gate is skipped." << std::endl;
      return input;
    }
  }

  fmi::bv model_miter (fmi::default_solver& solver
      , const fmi::bv& lhs
      , const fmi::bv& rhs)
  {
    assert( !solver.empty() && "No solver specified." );
    assert( fmi::bitsize( lhs ) == fmi::bitsize( rhs ) );

    return boost::fusion::at_c<0>( fmi::generate (solver, fmi::_0 != fmi::_1, lhs, rhs ) );
  }

  static void constraintConstantInputs (fmi::default_solver& solver, circuit const& circ, fmi::bv const& input)
  {
    vector < constant > const& consts = circ.constants();
    for (vector < constant >::const_iterator iter = consts.begin(); iter != consts.end(); ++iter)
    {
      constant const& c = *iter;
      if (c)
      {
        unsigned pos = distance ( consts.begin(), iter );

        fmi::bv v = fmi::extract ( solver, input, pos );
        fmi::assertion (solver,
            fmi::_0 == fmi::_1, 
            v, fmi::make_bin_constant (solver, *c ? "1" : "0" ) );
      }
    }
  }

  static map<string, string> get_input_default_mapping (circuit const& spec, circuit const& impl)
  {
    std::vector<std::string> const& specIns = spec.inputs();
    std::vector<std::string> const& implIns = impl.inputs();

    std::vector < constant > const& specConsts = spec.constants ();
    std::vector < constant > const& implConsts = impl.constants ();

    map  < string, string > mapping;
    for (unsigned i = 0; i < specIns.size(); i++)
    {
      if (!specConsts[i])
      {
        std::string const& specVar = specIns[i];

        std::vector<std::string>::const_iterator iter = find (implIns.begin(), implIns.end(), specVar);

        // FIXME iter == end()??
        if (iter != implIns.end())
        {
          unsigned pos = distance ( implIns.begin(), iter );

          if (!implConsts[pos])
          {
            mapping.insert (make_pair (specVar, specVar));
          }
        }
        else
        {
          mapping.insert (make_pair (specVar, implIns[i]));
        }
      }
    }

    return mapping;
  }

  static map<string, string> get_output_default_mapping (circuit const& spec, circuit const& impl)
  {
    std::vector<std::string> const& specOuts = spec.outputs();
    std::vector<std::string> const& implOuts = impl.outputs();

    std::vector < bool > const& specGarbages = spec.garbage ();
    std::vector < bool > const& implGarbages = impl.garbage ();

    map  < string, string > mapping;
    for (unsigned i = 0; i < specOuts.size(); i++)
    {
      if (!specGarbages[i])
      {
        std::string const& specVar = specOuts[i];

        std::vector<std::string>::const_iterator iter = find (implOuts.begin(), implOuts.end(), specVar);

        // FIXME iter == end()??
        if (iter != implOuts.end())
        {
          unsigned pos = distance ( implOuts.begin(), iter );

          if (!implGarbages[pos])
          {
            mapping.insert (make_pair (specVar, specVar));
          }
        }
        else
        {
          mapping.insert (make_pair (specVar, implOuts[i]));
        }
      }
    }

    return mapping;
  }


  // 
  // 0 -> input
  // 1 -> spec output 
  tr1::tuple<fmi::bv, fmi::bv> build_ec_model( const circuit& spec
      , const circuit& impl
      , properties::ptr settings
      , properties::ptr statistics
      , fmi::default_solver& solver )
  {
    fmi::bv specInput = fmi::new_variable( solver, spec.lines() );
    fmi::bv implInput = fmi::new_variable (solver, impl.lines() );

    constraintConstantInputs ( solver, spec, specInput );
    constraintConstantInputs ( solver, impl, implInput );

    fmi::bv lines = specInput;
    foreach ( const gate& aGate, spec )
    {
      lines = binary::model_gate( solver, aGate, lines );
    }
    fmi::bv specOutput = lines;

    lines = implInput; 
    foreach ( const gate& aGate, impl )
    {
      lines = binary::model_gate( solver, aGate, lines ); 
    }
    fmi::bv implOutput = lines;

    map<string,string> inMapping  = get<map<string,string> > (settings, "input_mapping",  map<string,string>());
    map<string,string> outMapping = get<map<string,string> > (settings, "output_mapping", map<string,string>());

    if (inMapping.empty())
      inMapping = get_input_default_mapping ( spec, impl );
    if (outMapping.empty())
      outMapping  = get_output_default_mapping ( spec, impl ); 

    vector<string> const& specInputs = spec.inputs();
    vector<string> const& implInputs = impl.inputs();
    vector<string> const& specOutputs = spec.outputs();
    vector<string> const& implOutputs = impl.outputs();

    for (map<string,string>::const_iterator iter = inMapping.begin(); iter != inMapping.end(); ++iter)
    {
      string const& specIn = iter->first;
      string const& implIn = iter->second;

      unsigned specPos = distance ( specInputs.begin(), find ( specInputs.begin(), specInputs.end(), specIn) );
      unsigned implPos = distance ( implInputs.begin(), find ( implInputs.begin(), implInputs.end(), implIn) );

      fmi::bv implVar = fmi::extract (solver, implInput, specPos);
      fmi::bv specVar = fmi::extract (solver, specInput, implPos);

      fmi::assertion ( solver, fmi::_0 == fmi::_1, implVar, specVar);
    }

    pair < vector < fmi::bv >, vector < fmi::bv > > miter;
    for (map<string,string>::const_iterator iter = outMapping.begin(); iter != outMapping.end(); ++iter)
    {
      string const& specOut = iter->first;
      string const& implOut = iter->second;

      unsigned specPos = distance ( specOutputs.begin(), find ( specOutputs.begin(), specOutputs.end(), specOut) );
      unsigned implPos = distance ( implOutputs.begin(), find ( implOutputs.begin(), implOutputs.end(), implOut) );

      fmi::bv implVar = fmi::extract (solver, implOutput, specPos);
      fmi::bv specVar = fmi::extract (solver, specOutput, implPos);

      miter.first.push_back (implVar);
      miter.second.push_back (specVar);
    }


    fmi::bv lhs = fmi::concat (solver, miter.first);
    fmi::bv rhs = fmi::concat (solver, miter.second);

    fmi::assertion( solver, model_miter( solver, lhs, rhs ) );

    return tr1::tuple<fmi::bv, fmi::bv> ( implInput, specOutput );
  }


  bool equivalence_check ( const circuit& spec
      , const circuit& impl
      , properties::ptr settings
      , properties::ptr statistics ) 

  {
    timer<properties_timer> t;

    if ( statistics )
    {
      properties_timer rt( statistics );
      t.start( rt );
    }

    fmi::default_solver solver  = fmi::get_solver_instance ( get<string>( settings, "solver", "MiniSAT" ) );
    unsigned max_counterexample = get<unsigned> (settings, "max_counterexample", 10u);

    fmi::bv input, specOutput;
    tr1::tie( input, specOutput ) = build_ec_model( spec, impl, settings, statistics, solver );

    fmi::solve_result s_result;

    vector<fmi::bv> vars;
    vars.push_back (input);
    vars.push_back (specOutput);

    counterexample cex;
    do 
    {
      s_result = fmi::solve (solver);

      // equivalent or no more countereaxmples exist 
      if (s_result == fmi::UNSAT) break;

      vector<boost::dynamic_bitset<> > assignments = fmi::get_assignment_vector( solver, vars );
      fmi::add_blocking_clause( solver, input, assignments[0] );

      cex.push_back ( make_pair( assignments[0], assignments[1] ) );
    } while ( s_result != fmi::UNSAT && cex.size() < max_counterexample);

    statistics->set ("equivalent", cex.empty() ? true : false);
    statistics->set ("counterexample", cex);
    return true;
  }

  equivalence_func equivalence_check_func (properties::ptr settings, properties::ptr statistics )
  {
    equivalence_func f = boost::bind( equivalence_check, _1, _2, settings, statistics );
    f.init ( settings, statistics );
    return f;
  }
}

