/* RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
 * Copyright (C) 2009-2010  The RevKit Developers <revkit@informatik.uni-bremen.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "read_realization.hpp"

#include <fstream>
#include <iostream>

#include "revlib_parser.hpp"
#include "../target_tags.hpp"
#include "../functions/add_gates.hpp"

namespace revkit
{

  ////////////////////////////// class circuit_processor
  class circuit_processor::priv
  {
  public:
    priv( circuit& c ) : circ( c ) {}

    circuit& circ;
  };

  circuit_processor::circuit_processor( circuit& circ )
    : revlib_processor(), d( new priv( circ ) )
  {
  }

  circuit_processor::~circuit_processor()
  {
    delete d;
  }

  void circuit_processor::on_numvars( unsigned numvars ) const
  {
    d->circ.set_lines( numvars );
  }

  void circuit_processor::on_inputs( std::vector<std::string>::const_iterator first, std::vector<std::string>::const_iterator last ) const
  {
    std::vector<std::string> inputs( first, last );
    d->circ.set_inputs( inputs );
  }

  void circuit_processor::on_outputs( std::vector<std::string>::const_iterator first, std::vector<std::string>::const_iterator last ) const
  {
    std::vector<std::string> outputs( first, last );
    d->circ.set_outputs( outputs );
  }

  void circuit_processor::on_constants( std::vector<constant>::const_iterator first, std::vector<constant>::const_iterator last ) const
  {
    std::vector<constant> constants( first, last );
    d->circ.set_constants( constants );
  }

  void circuit_processor::on_garbage( std::vector<bool>::const_iterator first, std::vector<bool>::const_iterator last ) const
  {
    std::vector<bool> garbage( first, last );
    d->circ.set_garbage( garbage );
  }

  void circuit_processor::on_gate( const boost::any& target_type, const std::vector<unsigned>& line_indices ) const
  {
    gate::line_container controls;

    if ( is_type<toffoli_tag>( target_type ) )
    {
      assert( line_indices.size() > 0 );
      std::copy( line_indices.begin(), line_indices.end() - 1, std::insert_iterator<gate::line_container>( controls, controls.begin() ) );
      append_toffoli( d->circ, controls, line_indices.back() );
    }
    else if ( is_type<fredkin_tag>( target_type ) )
    {
      assert( line_indices.size() > 1 );
      std::copy( line_indices.begin(), line_indices.end() - 2, std::insert_iterator<gate::line_container>( controls, controls.end() ) );
      append_fredkin( d->circ, controls, *( line_indices.end() - 2 ), *( line_indices.end() - 1 ) );      
    }
    else if ( is_type<peres_tag>( target_type ) )
    {
      assert( line_indices.size() == 3 );
      append_peres( d->circ, line_indices.at( 0 ), line_indices.at( 1 ), line_indices.at( 2 ) );
    }
    else if ( is_type<v_tag>( target_type ) )
    {
      assert( line_indices.size() == 2 );
      append_v( d->circ, line_indices.at( 0 ), line_indices.at( 1 ) );
    }
    else if ( is_type<vplus_tag>( target_type ) )
    {
      assert( line_indices.size() == 2 );
      append_vplus( d->circ, line_indices.at( 0 ), line_indices.at( 1 ) );
    }
  }

  bool read_realization( circuit& circ, std::istream& in, std::string* error )
  {
    circuit_processor processor( circ );

    return revlib_parser( in, processor, error );
  }

  bool read_realization( circuit& circ, const std::string& filename, std::string* error )
  {
    std::ifstream is;
    is.open( filename.c_str(), std::ifstream::in );

    if ( !is.good() )
    {
      if ( error )
      {
        *error = "Cannot open " + filename;
      }
      return false;
    }

    return read_realization( circ, is, error );
  }

}
