/* RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
 * Copyright (C) 2009-2010  The RevKit Developers <revkit@informatik.uni-bremen.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "revlib_parser.hpp"

#include <iostream>
#include <map>

#include <boost/algorithm/string/join.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/lexical_cast.hpp>

#include "../target_tags.hpp"
#include "revlib_processor.hpp"

namespace revkit
{

  struct transform_to_garbage
  {
    bool operator()( const char& c ) const
    {
      return c == '1';
    }
  };

  struct transform_to_constants
  {
    constant operator()( const char& c ) const
    {
      switch ( c ) {
      case '-':
        return constant();
        break;

      case '0':
      case '1':
        return constant( c == '1' );
        break;

      default:
        assert( false );
        return constant();
        break;
      }
    }
  };

  // CODE move to utils
  template<typename Map>
  struct map_access_functor
  {
    explicit map_access_functor( Map& map )
      : _map( map )
    {
    }

    const typename Map::mapped_type& operator()( const typename Map::key_type& key ) const
    {
      return _map.find( key )->second;
    }

  private:
    Map& _map;
  };

  template<typename Map>
  map_access_functor<Map> make_map_access_functor( Map& map )
  {
    return map_access_functor<Map>( map );
  }

  bool revlib_parser( std::istream& in, revlib_processor& reader, std::string* error )
  {
    std::string line;

    unsigned numvars = 0;
    unsigned truth_table_index = 0;
    std::map<std::string, unsigned> variable_indices;

    while ( in.good() && getline( in, line ) )
    {
      boost::algorithm::trim( line );

      /* skip empty lines */
      if ( !line.size() )
      {
        continue;
      }

      std::vector<std::string> params;
      boost::algorithm::split( params, line, boost::algorithm::is_any_of( " " ) );
      
      /* It is possible that there are empty elements in params,
         e.g. when line contains two spaces between identifiers instead of one.
         These should be removed. */
      std::vector<std::string>::iterator newEnd = std::remove( params.begin(), params.end(), "" );
      params.erase( newEnd, params.end() );

      /* By means of the first element we can determine the command */
      std::string command = params.front();
      params.erase( params.begin() );

      if ( command == "#" )
      {
        /* All parameters combined are considered as comment */
        reader.on_comment( boost::algorithm::join( params, " " ) );
      }
      else if ( command == ".version" )
      {
        /* All parameters combined are considered as version */
        reader.on_version( boost::algorithm::join( params, " " ) );
      }
      else if ( command == ".numvars" )
      {
        if ( params.size() != 1 )
        {
          if ( error )
          {
            *error = "Invalid number of parameters for .numvars command";
          }
          return false;
        }

        try
        {
          numvars = boost::lexical_cast<unsigned>( params.front() );
          reader.on_numvars( numvars );
        }
        catch ( boost::bad_lexical_cast& )
        {
          if ( error )
          {
            *error = "Invalid parameter for .numvars command";
          }
          return false;
        }
      }
      else if ( command == ".variables" )
      {
        if ( variable_indices.size() )
        {
          if ( error )
          {
            *error = "Second call of .variables command";
          }
          return false;
        }

        if ( params.size() != numvars )
        {
          if ( error )
          {
            *error = "Variable count does not fit numvars";
          }
          return false;
        }

        /* fill the index map later used when processing the gates */
        for ( std::vector<std::string>::const_iterator it = params.begin(); it != params.end(); ++it )
        {
          variable_indices.insert( std::make_pair( *it, it - params.begin() ) );
        }

        reader.on_variables( params.begin(), params.end() );
      }
      else if ( command == ".inputs" )
      {
        if ( params.size() != numvars )
        {
          if ( error )
          {
            *error = "Input count does not fit numvars";
          }
          return false;
        }

        reader.on_inputs( params.begin(), params.end() );
      }
      else if ( command == ".outputs" )
      {
        if ( params.size() != numvars )
        {
          if ( error )
          {
            *error = "Output count does not fit numvars";
          }
          return false;
        }

        reader.on_outputs( params.begin(), params.end() );
      }
      else if ( command == ".constants" )
      {
        if ( params.size() != 1 || params.front().size() != numvars )
        {
          if ( error )
          {
            *error = "Constant count does not fit numvars";
          }
          return false;
        }

        std::vector<constant> constants( numvars );
        std::transform( params.front().begin(), params.front().end(), constants.begin(), transform_to_constants() );

        reader.on_constants( constants.begin(), constants.end() );
      }
      else if ( command == ".garbage" )
      {
        if ( params.size() != 1 || params.front().size() != numvars )
        {
          if ( error )
          {
            *error = "Garbage count does not fit numvars";
          }
          return false;
        }

        std::vector<bool> garbage( numvars );
        std::transform( params.front().begin(), params.front().end(), garbage.begin(), transform_to_garbage() );

        reader.on_garbage( garbage.begin(), garbage.end() );
      }
      else if ( command == ".begin" )
      {
        if ( params.size() != 0 )
        {
          if ( error )
          {
            *error = "Wrong number of parameters for .begin command";
          }
          return false;
        }

        reader.on_begin();
      }
      else if ( command == ".end" )
      {
        if ( params.size() != 0 )
        {
          if ( error )
          {
            *error = "Wrong number of parameters for .end command";
          }
          return false;
        }

        reader.on_end();
      }
      else
      {
        if ( command[0] == '-' || command[0] == '1' || command[0] == '0' )
        {
          // truth table line
          if ( params.size() )
          {
            if ( error )
            {
              *error = "Params in truth table line";
            }
            return false;
          }

          std::vector<boost::optional<bool> > cube( command.size() );
          std::transform( command.begin(), command.end(), cube.begin(), transform_to_constants() );

          reader.on_truth_table_line( truth_table_index++, cube.begin(), cube.end() );
        }
        else
        {
          // gate
          std::vector<unsigned> line_indices( params.size() );
          std::transform( params.begin(), params.end(), line_indices.begin(), make_map_access_functor( variable_indices ) );

          boost::any gate_type;

          switch ( command[0] )
          {
          case 't':
            gate_type = toffoli_tag();
            break;

          case 'p':
            gate_type = peres_tag();
            break;

          case 'f':
            gate_type = fredkin_tag();
            break;

          case 'v':
            if ( command == "v+" )
            {
              gate_type = vplus_tag();
            }
            else if ( command == "v" )
            {
              gate_type = v_tag();
            }
            else
            {
              if ( error )
              {
                *error = "unknown gate command: " + command;
              }
              return false;
            }
            break;

          default:
            if ( error )
            {
              *error = "unknown gate command: " + command;
            }
          }

          reader.on_gate( gate_type, line_indices );
        }
      }
    }

    return true;
  }

}
