/* RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
 * Copyright (C) 2009-2010  The RevKit Developers <revkit@informatik.uni-bremen.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "write_blif.hpp"

#include <iterator>

#include <boost/assign/std/vector.hpp>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <boost/iterator/counting_iterator.hpp>
#include <boost/iterator/zip_iterator.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/tuple/tuple.hpp>

#include "../circuit.hpp"
#include "../target_tags.hpp"

using namespace boost::assign;

namespace revkit
{

  struct is_primary
  {
    template<typename T>
    bool operator()( const T& tuple ) const
    {
      return !boost::get<1>( tuple );
    }
  };

  template<int N, typename T>
  inline
  const typename boost::tuples::element<N,T>::type& getN( const T& t )
  {
    return boost::tuples::get<N>( t );
  }

  struct signal_name
  {
    signal_name( unsigned& tmp_signal, const std::string& tmp_signal_name )
      : tmp_signal( tmp_signal ),
        tmp_signal_name( tmp_signal_name ) {}

    template<typename T>
    std::string operator()( const T& tuple ) const
    {
      if ( !boost::get<1>( tuple ) )
      {
        return boost::get<0>( tuple );
      }
      else
      {
        return boost::str( boost::format( "%s%d" ) % tmp_signal_name % tmp_signal++ );
      }
    }

  private:
    unsigned& tmp_signal;
    const std::string& tmp_signal_name;
  };

  write_blif_settings::write_blif_settings()
    : tmp_signal_name( "tmp" )
  {
  }

  void write_blif_settings::operator()( std::ostream& os, const gate& g ) const
  {
    unsigned num_controls = std::distance( g.begin_controls(), g.end_controls() );

    if ( is_toffoli( g ) )
    {
      for ( unsigned j = 0; j < num_controls; ++j )
      {
        os << "1";

        for ( unsigned k = 0; k < j; ++k )
        {
          os << "-";
        }

        os << "0";

        for ( unsigned k = j + 1; k < num_controls; ++k )
        {
          os << "-";
        }

        os << " 1" << std::endl;
      }

      os << "0" << std::string( num_controls, '1' ) << " 1" << std::endl;
    }
    else if ( is_fredkin( g ) )
    {
      std::vector<std::string> not_equal;
      not_equal += "01","10";

      for ( unsigned i = 0; i < 2; ++i )
      {
        for ( unsigned j = 0; j < num_controls; ++j )
        {
          os << not_equal.at( i );

          for ( unsigned k = 0; k < j; ++k )
          {
            os << "-";
          }
          
          os << "0";

          for ( unsigned k = j + 1; k < num_controls; ++k )
          {
            os << "-";
          }

          os << " " << not_equal.at( i ) << std::endl;
        }

        os << not_equal.at( i ) << std::string( num_controls, '1' ) << " " << not_equal.at( 1 - i ) << std::endl;
      }

      os << "11" << std::string( num_controls, '-' ) << " 11" << std::endl;
    }
    else if ( is_peres( g ) )
    {
      // peres has one control and two targets
      os << "001 01" << std::endl
         << "010 01" << std::endl
         << "011 10" << std::endl
         << "100 10" << std::endl
         << "101 11" << std::endl
         << "110 11" << std::endl;
    }
  }

  void write_blif( const circuit& circ, std::ostream& os, const write_blif_settings& settings )
  {
    std::vector<std::string> signals( circ.lines() );
    unsigned tmp_signal = 0;

    /* model name */
    std::string model_name = circ.circuit_name();
    if ( !model_name.size() )
    {
      model_name = "circuit";
    }

    os << ".model " << model_name << std::endl;

    /* zip inputs and outputs */
    std::vector<boost::tuple<std::string, constant> > inputs( circ.lines() );
    std::copy( boost::make_zip_iterator( boost::make_tuple( circ.inputs().begin(), circ.constants().begin() ) ),
               boost::make_zip_iterator( boost::make_tuple( circ.inputs().end(), circ.constants().end() ) ),
               inputs.begin() );

    typedef boost::tuple<std::string, bool> out_tuple;
    std::vector<out_tuple> outputs( circ.lines() );
    std::copy( boost::make_zip_iterator( boost::make_tuple( circ.outputs().begin(), circ.garbage().begin() ) ),
               boost::make_zip_iterator( boost::make_tuple( circ.outputs().end(), circ.garbage().end() ) ),
               outputs.begin() );

    /* inputs */
    os << ".inputs ";
    std::transform( boost::make_filter_iterator<is_primary>( inputs.begin(), inputs.end() ),
                    boost::make_filter_iterator<is_primary>( inputs.end(), inputs.end() ),
                    std::ostream_iterator<std::string>( os, " " ),
                    getN<0, boost::tuple<std::string, constant> > );
    os << std::endl;

    /* outputs */
    os << ".outputs ";
    std::transform( boost::make_filter_iterator<is_primary>( outputs.begin(), outputs.end() ),
                    boost::make_filter_iterator<is_primary>( outputs.end(), outputs.end() ),
                    std::ostream_iterator<std::string>( os, " " ),
                    getN<0, out_tuple> );
    os << std::endl;

    std::transform( inputs.begin(), inputs.end(), signals.begin(), signal_name( tmp_signal, settings.tmp_signal_name ) );

    /* constants */
    tmp_signal = 0;
    foreach ( const constant& c, circ.constants() )
    {
      if ( c )
      {
        os << ".names " << settings.tmp_signal_name << tmp_signal++ << std::endl
           << "- " << ( *c ? "1" : "0" ) << std::endl;
      }
    }

    foreach ( const gate& g, circ )
    {
      unsigned num_targets = std::distance( g.begin_targets(), g.end_targets() );

      os << ".names";

      for ( gate::const_iterator it = g.begin_targets(); it != g.end_targets(); ++it )
      {
        os << " " << signals.at( *it );
      }

      for ( gate::const_iterator it = g.begin_controls(); it != g.end_controls(); ++it )
      {
        os << " " << signals.at( *it );
      }

      for ( unsigned i = 0; i < num_targets; ++i ) {
        os << " " << settings.tmp_signal_name << ( tmp_signal + 1 );
      }
      os << std::endl;

      settings( os, g );

      for ( gate::const_iterator it = g.begin_targets(); it != g.end_targets(); ++it )
      {
        signals.at( *it ) = boost::str( boost::format( "%s%d" ) % settings.tmp_signal_name % tmp_signal++ );
      }
    }

    foreach ( const out_tuple& t, outputs )
    {
      if ( !boost::get<1>( t ) )
      {
        os << ".names "
           << signals.at( std::find( circ.outputs().begin(), circ.outputs().end(), boost::get<0>( t ) ) - circ.outputs().begin() )
           << " "
           << boost::get<0>( t ) << std::endl
           << "1 1" << std::endl;
      }
    }

    os << ".end" << std::endl;
  }

}
