/* RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
 * Copyright (C) 2009-2010  The RevKit Developers <revkit@informatik.uni-bremen.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file write_blif.hpp
 *
 * @brief Writes a circuit to a BLIF file
 */

#ifndef WRITE_BLIF_HPP
#define WRITE_BLIF_HPP

#include <iostream>

namespace revkit
{
  
  class circuit;
  class gate;

  /**
   * @brief Settings for write_blif
   *
   * @author RevKit
   * @since  1.0
   */
  struct write_blif_settings
  {

    /**
     * @brief Standard constructor
     *
     * Initializes default values
     *
     * @author RevKit
     * @since  1.0
     */
    write_blif_settings();

    /**
     * @brief Prefix for the auxiliary variables which are created by the algorithm
     *
     * Default value is \b tmp
     *
     * @author RevKit
     * @since  1.0
     */
    std::string tmp_signal_name;

    /**
     * @brief Operator for transforming the gates into BLIF code
     *
     * By convention the first input signals are for the target lines and then the control lines.
     * The number of output signals is the number of target lines.
     * Only the cubes have to be printed, not the \em .names declaration.
     *
     * This operator has to be overridden when new gate types should be supported.
     *
     * @param os Output stream, from write_blif
     * @param g The current g to be transformed
     *
     * @author RevKit
     * @since  1.0
     */
    void operator()( std::ostream& os, const gate& g ) const;

  };

  /**
   * @brief Writes a circuit to a BLIF file
   *
   * This function writes the circuit to a BLIF file.
   * Therefore, it is required that the circuit has names for all its inputs and outputs.
   *
   * @param circ Circuit
   * @param os Output stream where to write the result to. Default is \b STDOUT.
   * @param settings Settings
   *
   * @author RevKit
   * @since  1.0
   */
  void write_blif( const circuit& circ, std::ostream& os = std::cout, const write_blif_settings& settings = write_blif_settings() );

}

#endif /* WRITE_BLIF_HPP */
