/* RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
 * Copyright (C) 2009-2010  The RevKit Developers <revkit@informatik.uni-bremen.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "write_realization.hpp"

#include <fstream>

#include <boost/algorithm/string/join.hpp>
#include <boost/algorithm/string/replace.hpp>
#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/iterator/counting_iterator.hpp>

#include "../circuit.hpp"
#include "../target_tags.hpp"

#define foreach BOOST_FOREACH

namespace revkit
{

  struct constant_to_char
  {
    char operator()( const constant& c )
    {
      return ( c ? ( *c ? '1' : '0' ) : '-' );
    }
  };

  struct garbage_to_char
  {
    char operator()( bool g )
    {
      return g ? '1' : '-';
    }
  };

  struct line_to_variable
  {
    std::string operator()( const gate::line& l ) const
    {
      return boost::str( boost::format( "x%d" ) % l );
    }
  };

  write_realization_settings::write_realization_settings()
    : version( "1.0" ),
      header( "This file has been generated using RevKit (www.revkit.org)" )
  {
  }

  void write_realization( const circuit& circ, std::ostream& os, const write_realization_settings& settings )
  {
    unsigned oldsize = 0;

    if ( settings.header.size() )
    {
      std::string header = settings.header;
      boost::algorithm::replace_all( header, "\n", "\n# " );
      os << "# " << header << std::endl;
    }
    
    os << ".version " << settings.version << std::endl
       << ".numvars " << circ.lines() << std::endl;

    std::vector<std::string> variables( circ.lines() );

    std::transform( boost::make_counting_iterator<unsigned>( 0 ),
                    boost::make_counting_iterator<unsigned>( circ.lines() ),
                    variables.begin(),
                    boost::bind( boost::str<char, std::char_traits<char>, std::allocator<char> >, boost::bind( &boost::format::operator%<unsigned>, boost::format( "x%d" ), _1 ) ) );

    std::vector<std::string> _inputs( circ.inputs().begin(), circ.inputs().end() );
    oldsize = _inputs.size();
    _inputs.resize( circ.lines() );
    if ( oldsize < circ.lines() )
    {
      std::transform( boost::make_counting_iterator<unsigned>( oldsize ),
                      boost::make_counting_iterator<unsigned>( circ.lines() ),
                      _inputs.begin(),
                      boost::bind( boost::str<char, std::char_traits<char>, std::allocator<char> >, boost::bind( &boost::format::operator%<unsigned>, boost::format( "i%d" ), _1 ) ) );
    }

    std::vector<std::string> _outputs( circ.outputs().begin(), circ.outputs().end() );
    oldsize = _outputs.size();
    _outputs.resize( circ.lines() );
    if ( oldsize < circ.lines() )
    {
      std::transform( boost::make_counting_iterator<unsigned>( oldsize ),
                      boost::make_counting_iterator<unsigned>( circ.lines() ),
                      _outputs.begin(),
                      boost::bind( boost::str<char, std::char_traits<char>, std::allocator<char> >, boost::bind( &boost::format::operator%<unsigned>, boost::format( "o%d" ), _1 ) ) );
    }

    os << ".variables " << boost::algorithm::join( variables, " " ) << std::endl
       << ".inputs " << boost::algorithm::join( _inputs, " " ) << std::endl
       << ".outputs " << boost::algorithm::join( _outputs, " " ) << std::endl;

    std::string _constants( circ.lines(), '-' );
    std::transform( circ.constants().begin(), circ.constants().end(), _constants.begin(), constant_to_char() );

    std::string _garbage( circ.lines(), '-' );
    std::transform( circ.garbage().begin(), circ.garbage().end(), _garbage.begin(), garbage_to_char() );

    os << ".constants " << _constants << std::endl
       << ".garbage " << _garbage << std::endl
       << ".begin" << std::endl;

    std::string cmd;

    foreach ( const gate& g, circ )
    {
      if ( is_toffoli( g ) )
      {
        cmd = boost::str( boost::format( "t%d" ) % g.size() );
      }
      else if ( is_fredkin( g ) )
      {
        cmd = boost::str( boost::format( "f%d" ) % g.size() );
      }
      else if ( is_peres( g ) )
      {
        cmd = "p";
      }
      else if ( is_v( g ) )
      {
        cmd = "v";
      }
      else if ( is_vplus( g ) )
      {
        cmd = "v+";
      }
      
      std::vector<std::string> lines;
      std::transform( g.begin_controls(), g.end_controls(), std::back_inserter( lines ), line_to_variable() );
      std::transform( g.begin_targets(), g.end_targets(), std::back_inserter( lines ), line_to_variable() );

      os << cmd << " " << boost::algorithm::join( lines, " " ) << std::endl;
    }

    os << ".end" << std::endl;
  }

  bool write_realization( const circuit& circ, const std::string& filename, const write_realization_settings& settings, std::string* error )
  {
    std::filebuf fb;
    if ( !fb.open( filename.c_str(), std::ios::out ) )
    {
      if ( error )
      {
        *error = "Cannot open " + filename;
      }
      return false;
    }

    std::ostream os( &fb );

    write_realization( circ, os, settings );

    fb.close();

    return true;
  }

}
