/* RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
 * Copyright (C) 2009-2010  The RevKit Developers <revkit@informatik.uni-bremen.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "costs.hpp"

#include <boost/tuple/tuple.hpp>

namespace revkit
{

  unsigned gate_costs::operator()( const circuit& circ ) const
  {
    return circ.num_gates();
  }

  unsigned line_costs::operator()( const circuit& circ ) const
  {
    return circ.lines();
  }

  unsigned quantum_costs::operator()( const gate& g, unsigned lines ) const
  {
    unsigned n = lines;
    unsigned c = std::distance( g.begin_controls(), g.end_controls() );
    unsigned e = n - c - 1; // empty lines

    switch ( c ) {
    case 0:
    case 1:
      return 1;
    case 2:
      return 5;
    case 3:
      return 13;
    case 4:
      return ( e >= 2 ) ? 26 : 29;
    case 5:
      if ( e >= 3 ) {
        return 38;
      } else if ( e >= 1 ) {
        return 52;
      } else {
        return 61;
      }
    case 6:
      if ( e >= 4 ) {
        return 50;
      } else if ( e >= 1 ) {
        return 80;
      } else {
        return 125;
      }
    case 7:
      if ( e >= 5 ) {
        return 62;
      } else if ( e >= 1 ) {
        return 100;
      } else {
        return 253;
      }
    case 8:
      if ( e >= 6 ) {
        return 74;
      } else if ( e >= 1 ) {
        return 128;
      } else {
        return 509;
      }
    case 9:
      if ( e >= 7 ) {
        return 86;
      } else if ( e >= 1 ) {
        return 152;
      } else {
        return 1021;
      }
    default:
      if ( e >= c - 2 ) {
        return 12 * c - 33;
      } else if ( e >= 1 ) {
        return 24 * c - 87;
      } else {
        return ( 1u << ( c + 1 ) ) - 3;
      }
    }
  }

  unsigned transistor_costs::operator()( const gate& g, unsigned lines ) const
  {
    return 8 * std::distance( g.begin_controls(), g.end_controls() );
  }

  struct costs_visitor : public boost::static_visitor<unsigned>
  {
    explicit costs_visitor( const circuit& circ ) : circ( circ ) {}

    unsigned operator()( const costs_by_circuit_func& f ) const
    {
      return f( circ );
    }

    unsigned operator()( const costs_by_gate_func& f ) const
    {
      unsigned sum = 0;
      foreach ( const gate& g, circ )
      {
        sum += f( g, circ.lines() );
      }
      return sum;
    }

  private:
    const circuit& circ;
  };

  unsigned costs( const circuit& circ, const cost_function& f )
  {
    return boost::apply_visitor( costs_visitor( circ ), f );
  }

}
