# RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
# Copyright (C) 2009-2010  The RevKit Developers <revkit@informatik.uni-bremen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from revkit import *

from PyQt4 import QtCore
from PyQt4 import QtGui

class CircuitLineItem( QtGui.QGraphicsItemGroup ):
    def __init__( self, index, width, parent = None ):
        QtGui.QGraphicsItemGroup.__init__( self, parent )

        # Create sub-lines
        x = 0
        for i in range( 0, width + 1 ):
            e_width = 15 if i == 0 or i == width else 30
            self.addToGroup( QtGui.QGraphicsLineItem( x, index * 30, x + e_width, index * 30 ) )
            x += e_width

    def apply_simulation_results( self, results ):
        map( lambda e, v: e.setPen( QtGui.QPen( QtCore.Qt.darkGreen if v else QtCore.Qt.red, 2 ) ), self.childItems(), results )

    def clear_simulation_results( self ):
        for item in self.childItems(): item.setPen( QtGui.QPen( QtCore.Qt.black, 0 ) )

class GateItem( QtGui.QGraphicsItemGroup ):
    def __init__( self, g, parent = None ):
        QtGui.QGraphicsItemGroup.__init__( self, parent )
        self.setFlag( QtGui.QGraphicsItem.ItemIsSelectable )

        l = control_lines( g )
        l.extend( target_lines( g ) )
        l.sort()

        if len( l ) > 1:
            circuitLine = QtGui.QGraphicsLineItem( 0, l[0] * 30, 0, l[-1] * 30, self )
            self.addToGroup( circuitLine )

        for c in control_lines( g ):
            control = QtGui.QGraphicsEllipseItem( -5, c * 30 - 5, 10, 10, self )
            control.setBrush( QtCore.Qt.black )
            self.addToGroup( control )

        for t in target_lines( g ):
            if g.type == gate_type.toffoli:
                target = QtGui.QGraphicsEllipseItem( -10, t * 30 - 10, 20, 20, self )
                targetLine = QtGui.QGraphicsLineItem( 0, t * 30 - 10, 0, t * 30 + 10, self )
                targetLine2 = QtGui.QGraphicsLineItem( -10, t * 30, 10, t * 30, self )
                self.addToGroup( target )
                self.addToGroup( targetLine )               
                self.addToGroup( targetLine2 )
            if g.type == gate_type.fredkin:
                crossTL_BR = QtGui.QGraphicsLineItem( -5, t * 30 - 5, 5, t * 30 + 5, self )
                crossTR_BL = QtGui.QGraphicsLineItem( 5, t * 30 - 5, -5, t * 30 + 5, self )
                self.addToGroup( crossTL_BR )
                self.addToGroup( crossTR_BL )
            if g.type == gate_type.v:
                target = QtGui.QGraphicsRectItem( -10, t * 30 - 10, 20, 20, self )
                target.setBrush( QtCore.Qt.white )
                text = QtGui.QGraphicsSimpleTextItem( "V", self )
                text.setPos( -4, t * 30 - 7 )
            if g.type == gate_type.vplus:
                target = QtGui.QGraphicsRectItem( -10, t * 30 - 10, 20, 20, self )
                target.setBrush( QtCore.Qt.white )
                text = QtGui.QGraphicsSimpleTextItem( "V+", self )
                text.setPos( -4, t * 30 - 7 )

    def itemChange( self, change, value ):
        if change == QtGui.QGraphicsItem.ItemSelectedChange:
            color = QtCore.Qt.red if value.toBool() else QtCore.Qt.black
                
            for item in filter( lambda x: type( x ) != QtGui.QGraphicsSimpleTextItem, self.childItems() ):
                item.setPen( QtGui.QPen( color ) )
                
            for item in filter( lambda x: x.brush().style() != QtCore.Qt.NoBrush and x.brush().color() != QtCore.Qt.white, filter( lambda x: issubclass( type( x ), QtGui.QAbstractGraphicsShapeItem ), self.childItems() ) ):
                item.setBrush( color )
        return QtGui.QGraphicsItemGroup.itemChange( self, change, value )

class CircuitView( QtGui.QGraphicsView ):
    def __init__( self, circ, parent = None ):
        QtGui.QGraphicsView.__init__( self, parent )

        self.circ = circ
        self.lines = []
        self.inputs = []
        self.outputs = []

        self.setScene( QtGui.QGraphicsScene( self ) )
        self.scene().setBackgroundBrush( QtCore.Qt.white )

        width = 30 * circ.num_gates
        
        for i in range(0, circ.lines):
            line = CircuitLineItem( i, circ.num_gates )
            self.lines.append( line )
            self.scene().addItem( line )
            self.inputs.append( self.addLineLabel( 0, i * 30, circ.inputs[i], QtCore.Qt.AlignRight, circ.constants[i] != None ) )
            self.outputs.append( self.addLineLabel( width, i * 30, circ.outputs[i], QtCore.Qt.AlignLeft, circ.garbage[i] ) )

        index = 0
        for g in circ.gates:
            gateItem = GateItem( g )
            gateItem.setPos( index * 30 + 15, 0 )
            self.scene().addItem( gateItem )
            index += 1
            
    def addLineLabel(self, x, y, text, align, color):
        textItem = self.scene().addText( text )
        textItem.setPlainText( text )

        if align == QtCore.Qt.AlignRight:
            x -= textItem.boundingRect().width()

        textItem.setPos( x, y - 12 )

        if color:
            textItem.setDefaultTextColor( QtCore.Qt.red )

        return textItem

    def simulate( self, inp ):
        if len( inp ) != self.circ.lines:
            return

        line_values = dict()
        for i in range( 0, len( inp ) ):
            line_values[i] = [not not inp[i]]

        def step_result_f( g, result ):
            for i in range( 0, len( result ) ):
                line_values[i].append( result[i] )

        output = []
        simple_simulation( output, self.circ, inp, step_result = step_result_f )

        # Display results
        for i in range( 0, len( inp ) ):
            self.inputs[i].setPlainText( self.circ.inputs[i] + " = " + str( int( line_values[i][0] ) ) )
            self.inputs[i].setPos( -self.inputs[i].boundingRect().width(), self.inputs[i].y() )
            self.outputs[i].setPlainText( self.circ.outputs[i] + " = " + str( int( line_values[i][-1] ) ) )
            self.lines[i].apply_simulation_results( line_values[i] )

    def clear_simulation( self ):
        for i in range( 0, self.circ.lines ):
            self.inputs[i].setPlainText( self.circ.inputs[i] )
            self.inputs[i].setPos( -self.inputs[i].boundingRect().width(), self.inputs[i].y() )
            self.outputs[i].setPlainText( self.circ.outputs[i] )
            self.lines[i].clear_simulation_results()

    def wheelEvent( self, event ):
        factor = 1.2
        if event.delta() < 0:
            factor = 1.0 / factor
        self.scale( factor, factor )


        

