#!/usr/bin/python

# RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
# Copyright (C) 2009-2010  The RevKit Developers <revkit@informatik.uni-bremen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, sys
from revkit import *
from revkitui import *

class InfoDialog( QtGui.QDialog ):
    def __init__( self, circ, parent ):
        QtGui.QDialog.__init__( self, parent, QtCore.Qt.Dialog )

        self.setWindowTitle( 'Circuit details' )
        self.resize( 20, 20 )

        self.setLayout( QtGui.QVBoxLayout() )
        self.layout().setMargin( 0 )

        widget = QtGui.QWidget()

        layout = QtGui.QFormLayout()
        widget.setLayout( layout )
        layout.addRow( QtGui.QLabel( 'Gate count:', widget ), QtGui.QLabel( str( costs( circ, gate_costs() ) ), widget ) )
        layout.addRow( QtGui.QLabel( 'Line count:', widget ), QtGui.QLabel( str( costs( circ, line_costs() ) ), widget ) )
        layout.addRow( QtGui.QLabel( 'Quantum cost:', widget ), QtGui.QLabel( str( costs( circ, quantum_costs() ) ), widget ) )
        layout.addRow( QtGui.QLabel( 'Transistor cost:', widget ), QtGui.QLabel( str( costs( circ, transistor_costs() ) ), widget ) )

        widget2 = QtGui.QWidget()
        widget2.setLayout( QtGui.QHBoxLayout() )
        widget2.layout().addStretch()
        button = QtGui.QPushButton( 'Close' )
        button.setAutoDefault( True )
        widget2.layout().addWidget( button ) 

        self.connect( button, QtCore.SIGNAL( 'clicked()' ), self.close )

        self.layout().addWidget( widget )
        self.layout().addWidget( widget2 )

class AboutDialog( QtGui.QDialog ):
    def __init__( self, parent ):
        QtGui.QDialog.__init__( self, parent, QtCore.Qt.Dialog )

        self.setWindowTitle( 'About RevKit Viewer' )
        self.resize( 20, 20 )

        self.setLayout( QtGui.QVBoxLayout() )
        self.layout().setMargin( 0 )

        widget2 = QtGui.QWidget()
        widget2.setLayout( QtGui.QHBoxLayout() )
        widget2.layout().addStretch()
        button = QtGui.QPushButton( 'Close' )
        button.setAutoDefault( True )
        widget2.layout().addWidget( button ) 

        self.connect( button, QtCore.SIGNAL( 'clicked()' ), self.close )

        self.layout().addWidget( QtGui.QLabel( '(c) 2009-2010 by the RevKit Developers' ) )
        self.layout().addWidget( widget2 )

class Viewer( QtGui.QMainWindow ):
    def __init__( self ):
        QtGui.QMainWindow.__init__( self )

        self.setWindowTitle( "RevKit Viewer" )

        self.setupActions()
        self.setupMenuBar()
        self.setupToolBar()
        self.statusBar()

        # CircuitView
        circ = circuit( 0 )
        self.view = CircuitView( circ, self )
        self.setCentralWidget( self.view )

        self.resize( 400, 300 )

    def setupActions( self ):
        path = os.path.realpath( os.path.abspath( __file__ ) )
        path = path.replace( os.path.basename( __file__ ), 'icons/' )

        self.openAction = QtGui.QAction( QtGui.QIcon( path + 'document-open.png' ), '&Open...', self )
        self.openAction.setStatusTip( 'Open a circuit realization in RevLib format' )
        self.imageAction = QtGui.QAction( QtGui.QIcon( path + 'image-x-generic.png' ), 'Save as &Image...', self )
        self.imageAction.setStatusTip( 'Saves the circuit as an image file (PNG or JPG)' )
        self.exitAction = QtGui.QAction( QtGui.QIcon( path + 'application-exit.png' ), '&Quit', self )
        self.exitAction.setStatusTip( 'Quits the program' )

        self.infoAction = QtGui.QAction( QtGui.QIcon( path + 'dialog-information.png' ), '&Circuit details', self )
        self.infoAction.setStatusTip( 'Opens a dialog with circuit information' )
        self.specAction = QtGui.QAction( QtGui.QIcon( path + 'view-form-table.png' ), '&View truth table', self )
        self.specAction.setStatusTip( 'Displays the full truth table of the circuit, obtained by simulation' )
        self.partialAction = QtGui.QAction( QtGui.QIcon( path + 'view-form-table.png' ), '&View partial truth table', self )
        self.partialAction.setStatusTip( 'Displays a truth table only for non-constant and non-garbage signals' )

        self.aboutAction = QtGui.QAction( QtGui.QIcon( path + 'help-about.png' ), '&About', self )
        self.aboutAction.setStatusTip( 'Displays information about the RevKit Viewer' )

        self.connect( self.openAction, QtCore.SIGNAL( 'triggered()' ), self.open )
        self.connect( self.imageAction, QtCore.SIGNAL( 'triggered()' ), self.saveImage )
        self.connect( self.exitAction, QtCore.SIGNAL( 'triggered()' ), QtCore.SLOT( 'close()' ) )

        self.connect( self.infoAction, QtCore.SIGNAL( 'triggered()' ), self.info )
        self.connect( self.specAction, QtCore.SIGNAL( 'triggered()' ), self.truthTable )
        self.connect( self.partialAction, QtCore.SIGNAL( 'triggered()' ), self.partialTable )

        self.connect( self.aboutAction, QtCore.SIGNAL( 'triggered()' ), self.about )
        
    def setupMenuBar( self ):
        menubar = self.menuBar()

        file = menubar.addMenu( '&File' )
        file.addAction( self.openAction )
        file.addAction( self.imageAction )
        file.addSeparator()
        file.addAction( self.exitAction )

        view = menubar.addMenu( '&View' )
        view.addAction( self.infoAction )
        view.addSeparator()
        view.addAction( self.specAction )
        view.addAction( self.partialAction )
        
        help = menubar.addMenu( '&Help' )
        help.addAction( self.aboutAction )

    def setupToolBar( self ):
        toolbar = self.addToolBar( 'Main' )
        toolbar.setIconSize( QtCore.QSize( 32, 32 ) )

        toolbar.addAction( self.openAction )
        toolbar.addAction( self.imageAction )
        toolbar.addSeparator()
        toolbar.addAction( self.infoAction )
        toolbar.addAction( self.partialAction )

    def open( self ):
        filename = str( QtGui.QFileDialog.getOpenFileName( self, 'Open Realization', '', 'RevLib Realization (*.real)' ) )
        if len( filename ):
            self.openCircuit( filename )

    def openCircuit( self, filename ):
        circ = circuit()
        read_realization( circ, filename )

        self.centralWidget().close()
        self.view = CircuitView( circ, self )
        self.setCentralWidget( self.view )

    def saveImage( self ):
        filename = QtGui.QFileDialog.getSaveFileName( self, 'Save as Image', '', 'PNG image (*.png);;JPG image (*.jpg)' )
        if not filename.isEmpty():
            scene = self.view.scene()
            pixmap = QtGui.QPixmap( scene.width(), scene.height() )
            painter = QtGui.QPainter( pixmap )
            scene.render( painter )
            pixmap.save( filename )
            painter.end()

    def info( self ):
        dialog = InfoDialog( self.view.circ, self )        
        dialog.exec_()

    def truthTable( self ):
        dialog = QtGui.QDialog( self, QtCore.Qt.Dialog )
        dialog.setWindowTitle( 'Truth Table' )

        spec = binary_truth_table()
        circuit_to_truth_table( self.view.circ, spec )

        n = self.view.circ.lines

        table = QtGui.QTableWidget( 2 ** n, 2 * n, dialog )
        table.setHorizontalHeaderLabels( self.view.circ.inputs + self.view.circ.outputs )
        table.setVerticalHeaderLabels( map( str, range( 0, 2 ** n ) ) )
        table.setAlternatingRowColors( True )
        table.setShowGrid( False )

        row = 0
        for entry in spec.entries:

            valid = True
            for c in range( 0, n ):
                if not self.view.circ.constants[c] is None and entry[0][c] != self.view.circ.constants[c]:
                    valid = False
                    break
                
            
            for col in range( 0, 2 * n ):
                item = QtGui.QTableWidgetItem( '1' if ( entry[0] + entry[1] )[col] else '0' )
                flags = QtCore.Qt.ItemIsSelectable
                if valid and not ( col >= n and self.view.circ.garbage[col % n] ) and not ( col < n and not self.view.circ.constants[col] is None ):
                    flags |= QtCore.Qt.ItemIsEnabled
                item.setFlags( flags )
                if col >= n and not self.view.circ.garbage[col % n]:
                    item.setBackground( QtCore.Qt.lightGray )
                table.setItem( row, col, item )
            row += 1

        table.resizeColumnsToContents()

        dialog.setLayout( QtGui.QVBoxLayout() )
        dialog.layout().setMargin( 0 )
        dialog.layout().addWidget( table )
        
        dialog.exec_()

    def partialTable( self ):
        dialog = QtGui.QDialog( self, QtCore.Qt.Dialog )
        dialog.setWindowTitle( 'Partial Truth Table' )

        spec = binary_truth_table()
        settings = properties()
        settings.set_bool( "partial", True )
        circuit_to_truth_table( self.view.circ, spec, py_partial_simulation_func( settings ) )

        n = len( filter( lambda x: x is None, self.view.circ.constants ) )
        m = len( filter( lambda x: not x, self.view.circ.garbage ) )

        table = QtGui.QTableWidget( 2 ** n, n + m, dialog )
        input_labels = map( lambda x: x[0], filter( lambda x: x[1] is None, map( lambda x, y: [x,y], self.view.circ.inputs, self.view.circ.constants ) ) )
        output_labels = map( lambda x: x[0], filter( lambda x: not x[1], map( lambda x, y: [x,y], self.view.circ.outputs, self.view.circ.garbage ) ) ) 
        table.setHorizontalHeaderLabels( input_labels + output_labels )
        table.setVerticalHeaderLabels( map( lambda x: "", range( 0, 2 ** n ) ) )
        table.setAlternatingRowColors( True )
        table.setShowGrid( False )

        row = 0
        for entry in spec.entries:           
            for col in range( 0, n + m ):
                item = QtGui.QTableWidgetItem( '1' if ( entry[0] + entry[1] )[col] else '0' )
                item.setFlags( QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled )
                if col >= n:
                    item.setBackground( QtCore.Qt.lightGray )
                table.setItem( row, col, item )
            row += 1

        table.resizeColumnsToContents()

        dialog.setLayout( QtGui.QVBoxLayout() )
        dialog.layout().setMargin( 0 )
        dialog.layout().addWidget( table )
        
        dialog.exec_()

    def about( self ):
        dialog = AboutDialog( self )        
        dialog.exec_()

if __name__ == '__main__':
    a = QtGui.QApplication([])

    w = Viewer()
    w.show()

    if len( sys.argv ) == 2:
        w.openCircuit( sys.argv[1] )

    sys.exit( a.exec_() )
